/////////////////////////////////////////////////////////////////////////
// Unit principale du projet ZnZTest
//-----------------------------------------------------------------------
// Copyright (c) 1997, Marcel Martin. Tous droits rservs.
/////////////////////////////////////////////////////////////////////////
unit ZnZMain;
{$A+,H+}

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, Menus;

type
  TZnZForm = class(TForm)
    Memo: TMemo; //Ou TRichEdit pour les trs grands nombres
    CtrlPanel: TPanel;
    MenuBtn: TBitBtn;
    ClearMemoBtn: TBitBtn;
    AbortBtn: TBitBtn;
    CloseBtn: TBitBtn;
    PopMenu: TPopupMenu;
    PopItm1: TMenuItem;
    PopItm2: TMenuItem;
    PopItm3: TMenuItem;
    PopItm4: TMenuItem;
    procedure AbortBtnClick(Sender: TObject);
    procedure ClearMemoBtnClick(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure MenuBtnClick(Sender: TObject);
    procedure PopItm1Click(Sender: TObject);
    procedure PopItm2Click(Sender: TObject);
    procedure PopItm3Click(Sender: TObject);
    procedure PopItm4Click(Sender: TObject);
    procedure Run1(Sender: TObject);
    procedure Run2(Sender: TObject);
    procedure Run3(Sender: TObject);
    procedure Run4(Sender: TObject);
    procedure Runx(Sender: TObject; Num: Integer);
  end;

var ZnZForm: TZnZForm;


implementation
//=======================================================================
{$I-,O+,Q-,R-}
//=======================================================================
{$IFDEF VER90}  //DELPHI 2
uses ZnZ_d2;
{$ELSE}
{$IFDEF VER100} //DELPHI 3
  {$C-,G-}
uses ZnZ_d3;
{$ENDIF}
{$ENDIF}

{$R *.DFM}

//=======================================================================
// Permet  certaines routines de ZnZ d'interagir avec le programme
//=======================================================================
procedure Idle;
begin
  Application.ProcessMessages;
end;

//=======================================================================
// Interrompt l'excution (ventuelle) de certaines routines de ZnZ
//=======================================================================
procedure TZnZForm.AbortBtnClick(Sender: TObject);
begin
  ZnZAbort;
end;

//=======================================================================
// Efface l'diteur
//=======================================================================
procedure TZnZForm.ClearMemoBtnClick(Sender: TObject);
begin
  Memo.Lines.Clear;
end;

//=======================================================================
// Referme la fiche (Quitte le programme)
//=======================================================================
procedure TZnZForm.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

//=======================================================================
// Cration de la fiche
//=======================================================================
procedure TZnZForm.FormCreate(Sender: TObject);
begin
  Application.HintColor := clYELLOW;
  //----- Neutralise le compactage automatique des nombres
  //PackedNumbers := false;
  //----- "Flushe" les nombres avant destruction
  //CryptoLevel := true;
end;

//////////////////////////////////////////////////////////////////////////
// Dmo 1
//////////////////////////////////////////////////////////////////////////

//=======================================================================
// Calcule 2^(N-1) mod N avec N alatoire de 256 bits
//-----------------------------------------------------------------------
// Avec un composant TRichEdit, pour diter des chanes de plus de 4094
// caractres, remplacez
//   "Memo.Lines.Add(S);"
// par
//   "Memo.Text := Memo.Text + S + #13#10;"
//=======================================================================
procedure TZnZForm.Run1(Sender: TObject);
  var
    N, E, X      : TNumber;
    BSize, Hw, c : Longint;
begin
  N := nil; E := nil; X := nil;
  try
    N := TNumber.Create; //Cre le nombre N
    E := TNumber.Create;
    X := TNumber.Create;
    //----- Pas de signe dans l'dition des chanes de caractres
    SignedStrs := false;
    with Memo.Lines do
    begin
      //----- Gnration d'un nombre alatoire
      BSize := 256;       //256 bits ~ 78 chiffres dcimaux
      Rndn(N,BSize,true); //"true" pour que la taille binaire = BSize
      N.Odd := true;      //Pour que le nombre soit impair
      //----- Conversion du nombre en chane dcimale et affichage
      Add('N =');
      Add(N.Str(10,3));
      //----- Affichage de la taille en base 10 du nombre
      Add('Taille = ' + IntToStr(N.DecimalSize) + ' chiffres');
      Add('');
      //----- E = N-1
      Movn(N,E); //Copie de N dans E
      Sub1(E);   //E := E - 1
      //----- X = 2
      Movi(2,X);
      //----- Calcul de X := 2^(N-1) mod N
      c := GetTickCount;
      PowMod(X,E,N);
      c := GetTickCount - c;
      Add('2^(N-1) mod N =');
      Add(X.Str(10,3));
      Add('Taille             = ' + IntToStr(X.DecimalSize) + ' chiffre(s)');
      Add('Taille (base 2^32) = ' + IntToStr(X.LongsCount));
      Add('Prcision          = ' + IntToStr(X.Precision));
      //----- Si X = 1 (Evnement rare: Proba ~ 2/N.BitsCount)
      if X.Equals1 then Add('N est pseudo-premier de base 2.');
      Add('');
      //----- Travail effectu (dpend de la taille et du poids de Hamming)
      Hw := Pred(E.HammingWeight);
      Dec(BSize);
      Add('Ce calcul a demand ' +
           IntToStr(BSize) + ' carrs, ' +
           IntToStr(Hw) + ' multiplications et ' +
           IntToStr(BSize+Hw) + ' divisions.');
      //----- Affichage du chronomtrage
      Add('Dure = ' + IntToStr(c) + ' ms.');
      Add('');
    end;
  finally
    //----- Destruction des nombres
    N.Free;
    E.Free;
    X.Free;
    //----- Restaure la valeur par dfaut
    SignedStrs := true;
  end;
end;

//////////////////////////////////////////////////////////////////////////
// Dmo 2
//////////////////////////////////////////////////////////////////////////

//=======================================================================
// Compare les dures d'excution de Sqrn et de FFTSqr
//=======================================================================
procedure TZnZForm.Run2(Sender: TObject);
  const CHECKMSG : array[Boolean] of string[30] =
                  ('Ce message n''apparatra jamais',
                   'OK');
  var
    X, Y        : TNumber;
    Len, c1, c2 : Longint;
begin
  X := nil;
  Y := nil;
  try
    with Memo.Lines do
    begin
      X := TNumber.Create; //Cre le nombre X
      //----- Gnration d'un nombre alatoire
      // Sur CYRIX 166+, avec 1 000 000 de bits le calcul du carr
      // "classique" prend environ 2 mn, FFT demande moins de 7 s.
      //--------------------------------------------------------------
      // Excutez cette routine avec la valeur ci-dessous, puis
      // augmentez la valeur d'un bit. Vous constaterez que la dure du
      // calcul pour FFT est un peu plus du double de la prcdente.
      // C'est tout  fait normal: A cause des vecteurs FFT, il y a des
      // seuils.
      //----- 2048 * 24 < 98304 <= 4096 * 24, le vecteur FFT est de
      //      dimension 4096 (en base 2^24)
      //----- 4096 * 24 < 98305 <= 8192 * 24, le vecteur FFT est de
      //      dimension 8192 (en base 2^24)
      Len := 98304;
      Rndn(X,Len,true);
      //----- Cre le nombre Y gal  X
      Y := TNumber.Copy(X);
      //----- Pour interrompre les calculs
      ZnZIdle(Idle,true);
      try
        //----- Calcul de X
        c1 := GetTickCount;
        Sqrn(X);
        c1 := GetTickCount - c1;
        //----- Calcul de Y
        c2 := GetTickCount;
        FFTSqr(Y);
        c2 := GetTickCount - c2;
      finally
        ZnZNoIdle; //Repasse en mode sans interruption
      end;
      //----- Affiche les chronomtrages
      Add('Dure Sqrn   = ' + IntToStr(c1) + ' ms.');
      Add('Dure FFTSqr = ' + IntToStr(c2) + ' ms.');
      //----- Vrification
      Add('> ' + CHECKMSG[Cmpn(X,Y) = 0]);
      Add('');
    end;
  finally
    //----- Destruction des nombres
    X.Free;
    Y.Free;
  end;
end;

//////////////////////////////////////////////////////////////////////////
// Dmo 3
//////////////////////////////////////////////////////////////////////////

//=======================================================================
// Empreintes numriques MD5, RIPEMD-128, SHA-1 et RIPEMD-160
//=======================================================================
procedure TZnZForm.Run3(Sender: TObject);
  var
  //S : WideString;  //Oui, a marche aussi, mais c'est plus long
  //S : ShortString; //Seulement avec les chanes de test 1 et 2.
    S : string;      //Equivalent  AnsiString
  //-------------------------------------------
  procedure Fingerprints;
    const
      Title : array[THFunc] of string[12] =
             ('MD5        =','RIPEMD-128 =',
              'SHA-1      =','RIPEMD-160 =');
    var
      X : TNumber;
      f : THFunc;
  begin
    Memo.Lines.Add('');
    X := TNumber.Create; //Cre le nombre X
    try
      //----- Itre sur MD5, RMD128, SHA1 et RMD160
      for f := Low(THFunc) to High(THFunc) do
      begin
        //----- Calcul de l'empreinte de S
        HString(X,S,f);
        //----- Affichage du l'empreinte
        Memo.Lines.Add(Title[f] + X.Str(16,8));
      end;
      Memo.Lines.Add('');
    finally
      //----- Destruction du nombre
      X.Free;
    end;
  end;
  //-------------------------------------------
begin
  //----- Chane 1
  S := 'abc';
  Memo.Lines.Add('S1 = "' + S + '"');
  Fingerprints;
  //----- Chane 2
  S := 'abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq';
  Memo.Lines.Add('S2 = "' + S + '"');
  Fingerprints;
  //----- Chane 3
  S := StringOfChar('a',1000000);
  Memo.Lines.Add('S3 = "aaa...a" (1 000 000 de caractres "a")');
  Fingerprints;
end;

//////////////////////////////////////////////////////////////////////////
// Dmo 4
//////////////////////////////////////////////////////////////////////////

//=======================================================================
// Gnrateur alatoire cryptographiquement sr BBS (Blum-Blum-Shub)
//-----------------------------------------------------------------------
// Les gnrateurs BBS sont les meilleurs gnrateurs pseudo-alatoires
// qui existent. Ils sont relativement lents mais ce sont les plus srs.
//-----------------------------------------------------------------------
// Cette routine se trouve dans les "programmes-exemples" de ZnZ.HLP
//=======================================================================
procedure TZnZForm.Run4(Sender: TObject);
  const
    NSIZE = 256;   //Taille du module
    XBITS = 10000; //Nombre de bits alatoires
    FNAME = 'BBS Modulus 256 bits.ZnZ'; //Adaptez selon NSIZE
  var
    S, N, X          : TNumber;
    BSize, c, BCount : Longint;
    Mask             : Word;
begin
  S := nil; N := nil; X := nil;
  try
    //----- Pas de signe dans l'dition des chanes de caractres
    SignedStrs := false;
    //----- Cration des nombres
    S := TNumber.Create; //S (Germe)
    N := TNumber.Create; //N (Module)
    X := TNumber.Create;
    with Memo.Lines do
    begin
      //--------------------------------------------------------
      // Si le fichier FNAME n'existe pas, le module est cr et
      // sauvegard, sinon il est charg depuis le fichier.
      //--------------------------------------------------------
      if not FileExists(FNAME) then
      begin
        //----- Calcule du module (NSIZE bits) du gnrateur
        ZnZIdle(Idle,false); //Pour interrompre les calculs
        try
          c := GetTickCount;
          //----- Facteur #1
          BSize := NSIZE shr 1;
          repeat
            //----- "true" assure que N.BitsCount = BSize
            Rndn(N,BSize,true);
            //----- Assure que (N*X).BitsCount = NSIZE
            N.Bits[BSize-2] := true;
            //----- N fortement pseudo-premier de bases 2, 3, 5 et 7
            //      N congru  3 modulo 4
            MakePrime(N,3,4,[2,3,5,7]);
          until N.BitsCount = BSize;
          //----- Facteur #2
          BSize := NSIZE - BSize;
          repeat
            Rndn(X,BSize,true);
            X.Bits[BSize-2] := true;
            MakePrime(X,3,4,[2,3,5,7]);
          until (X.BitsCount = BSize) and (Cmpn(N,X) <> 0);
          c := GetTickCount - c;
        finally
          ZnZNoIdle;
        end;
        //----- Calcul du module, N := N * X (N est un entier de Blum)
        Muln(N,X);
        X.Set0;
        //----- Sauvegarde du module
        N.SaveToFile(FNAME);
      end
      //----- Chargement du module
      else
      begin
        N.LoadFromfile(FNAME);
        c := 0;
      end;
      //----- Affichage du module
      Add('Module N (Entier de Blum) =');
      Add(N.Str(10,3));
      //----- Informations sur le nombre
      Add('Taille binaire   = ' + IntToStr(N.BitsCount));
      Add('Poids de Hamming = ' + IntToStr(N.HammingWeight));
      //----- Affichage du chronomtrage
      if c <> 0 then Add('Dure ' + IntToStr(c) + ' ms.');
      Add('');

      //===== Calcule le germe S[0] du gnrateur
      repeat
        Rndn(S,NSIZE+NSIZE,false); //S, alatoire d'au plus 2*NSize bits
        Modn(S,N);
      until Gcd1(S,N) and (Cmpi(S,1) > 0);
      //----- S[0] := Z mod N, avec (1 < Z < N) et (PGCD(Z,N) = 1)
      Sqrn(S);
      Modn(S,N);
      //----- Affichage du germe
      Add('S (Germe) =');
      Add(S.Str(10,3));
      Add('');

      //===== Calcule un nombre alatoire de XBITS bits
      c := GetTickCount;
      //----- BitCount, nombre de bits cryptographiquement srs
      Movi(NSIZE,X);
      BCount := Pred(X.BitsCount); //BitCount ~ Log2(Log2(N))
      if BCount > 16 then BCount := 16;
      X.Set0;
      //----- Mask := 111..11b (BitCount bits  1)
      Mask := Word(Pred(Longint(1) shl BCount));
      BSize := XBITS;
      //----- Si "PackedNumbers" alors "Expand" est une perte de temps
      if not PackedNumbers then X.Expand((XBITS+31) shr 5);
      //------------------------------------------------------------
      // L'expression prcdente est correcte. Toutefois, soyez trs
      // prudents lorsque vous utilisez "shr" avec 2 constantes.
      // Normalement $ffffffff shr 16 = $0000ffff, mais pour le
      // compilateur de DELPHI (qui semble confondre "shr" avec le
      // dcalage arithmtique "sar") $ffffffff shr 16 = $ffffffff.
      // Oui, ce bug est invraisemblable.
      //------------------------------------------------------------
      while BSize >= BCount do
      begin
        //----- Concatne X et les BitCount derniers bits de S
        Shls(X,BCount);
        X.Words[0] := X.Words[0] or (S.Words[0] and Mask);
        Dec(BSize,BCount);
        //----- Itre le gnrateur: S[i+1] := S[i] mod N
        Sqrn(S);
        Modn(S,N);
      end;
      //---- Reliquat si (XBITS modulo BitCount) <> 0
      if BSize > 0 then
      begin
        Shls(X,BSize);
        X.Words[0] := X.Words[0] or
                      (S.Words[0] and Word(Pred(Longint(1) shl BSize)));
      end;
      c := GetTickCount - c;
      //----- Affichage du nombre alatoire
      Add('X =');
      Add(X.Str(10,3));
      //----- Informations sur le nombre
      Add('Taille binaire   = ' + IntToStr(X.BitsCount));
      Add('Taille dcimale  = ' + IntToStr(X.DecimalSize));
      Add('Poids de Hamming = ' + IntToStr(X.HammingWeight));
      Add(IntToStr(BCount) + ' bits ont t retenus  chaque itration');
      //----- Affichage du chronomtrage
      Add('Dure ' + IntToStr(c) + ' ms.');
      Add('');
    end;
  finally
    //----- Destruction des nombres
    S.Free;
    N.Free;
    X.Free;
    //----- Restaure la valeur par dfaut
    SignedStrs := true;
  end;
end;

//////////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////////

//=======================================================================
// Ouvre le menu Popup
//=======================================================================
procedure TZnZForm.MenuBtnClick(Sender: TObject);
  var P : TPoint;
begin
  MenuBtn.Enabled := false;
  try
    GetCursorPos(P);
    PopMenu.Popup(P.X,P.Y);
  finally
    MenuBtn.Enabled := true;
  end;
end;

//=======================================================================
// Appelle une des 4 mthodes Run1, Run2, Run3 ou Run4
//=======================================================================
procedure TZnZForm.Runx(Sender: TObject; Num: Integer);
begin
  MenuBtn.Enabled := false;
  PopMenu.AutoPopup := false;
  try
    Memo.Lines.Add(#13#10'Dmo ' + IntToStr(Num));
    Memo.Lines.Add('------'#13#10);
    case Num of
      1: Run1(Sender);
      2: Run2(Sender);
      3: Run3(Sender);
      else Run4(Sender);
    end;
  finally
    MenuBtn.Enabled := true;
    PopMenu.AutoPopup := true;
  end;
end;

//=======================================================================
// Appelle la mthode Run1
//=======================================================================
procedure TZnZForm.PopItm1Click(Sender: TObject);
begin
  Runx(Sender,1);
end;

//=======================================================================
// Appelle la mthode Run2
//=======================================================================
procedure TZnZForm.PopItm2Click(Sender: TObject);
begin
  Runx(Sender,2);
end;

//=======================================================================
// Appelle la mthode Run3
//=======================================================================
procedure TZnZForm.PopItm3Click(Sender: TObject);
begin
  Runx(Sender,3);
end;

//=======================================================================
// Appelle la mthode Run4
//=======================================================================
procedure TZnZForm.PopItm4Click(Sender: TObject);
begin
  Runx(Sender,4);
end;

//////////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////////

{ La fonction suivante n'est pas utilise par ce projet, mais elle peut }
{ vous tre utile.                                                      }

//=======================================================================
// Test pour FFT
// -> Min et Max, tailles en octets
// <- Result = -1 si tout est correct
//           = indice de l'octet non valide si erreur
//-----------------------------------------------------------------------
// Cette routine teste si le rsultat d'un produit calcul avec FFT est
// correct. En effet, au-del d'une certaine taille, il n'est plus
// possible de comparer avec le rsultat des routines "classiques", cela
// pourrait prendre des heures...
// On calcule le produit (256**Min - 1)*(256**Max - 1). Le rsultat doit
// toujours avoir la structure suivante:
// Octet  [0]                = 1
// Octets [1..Min-1]         = 0
// Octets [Min..Max-1]       = 255  (Si Min <> Max, sinon n'existent pas)
// Octet  [Max]              = 254
// Octets [Max+1..Min+Max-1] = 255
//
// Bien sr le fait que cette fonction renvoie systmatiquement -1 ne
// prouve pas que les routines FFT de ZnZ soient correctes, mais c'est
// tout de mme rassurant, non ?
//=======================================================================
function FFTCheck(Min, Max: Longint): Longint;
  const XMSG = 'Fonction FFTCheck: (Min,Max) non valide';
  var X, Y : TNumber;
begin
  //----- Tests de validit
  if (0 > Min) or (Min > Max) or (Max > (MaxNumberPrecision*2)) then
    ZnZError(XMSG);
  X := TNumber.Create;
  try
    X.FullBits(Min shl 3);
    //----- Si Min = Max, teste FFTSqr
    if Min = Max then FFTSqr(X)
    else
    //----- Si Min < Max, teste FFTMul
    begin
      Y := TNumber.Create;
      try
        Y.FullBits(Max shl 3);
        FFTMul(X,Y);
      finally
        Y.Free;
      end;
    end;
    //----- Tests
    Result := 0;
    if X.Bytes[Result] <> $01 then Exit;
    for Result := 1 to Pred(Min) do
      if X.Bytes[Result] <> $00 then Exit;
    for Result := Min to Pred(Max) do
      if X.Bytes[Result] <> $ff then Exit;
    Result := Max;
    if X.Bytes[Result] <> $fe then Exit;
    for Result := Succ(Max) to Pred(X.BytesCount) do
      if X.Bytes[Result] <> $ff then Exit;
    //----- Tout est correct
    Result := -1;
  finally
    X.Free;
  end;
end;

/////////////////////////////////////////////////////////////////////////
end.
/////////////////////////////////////////////////////////////////////////

